//Copyright (C) 1998 John Meacham 
//john@foo.net

#include <stdlib.h>
#include <values.h>
#include <math.h>
#include <stdio.h>
#include <unistd.h>

#include "engine.h"
#include "multisection.h"
#include "primitive.h"

globals g;
object* head;  //first to object looses

list<lamp*> lamps;

#define MAX_DEPTH 5

void panic(char *c) {
   fprintf(stderr,"PANIC: %s\n",c);
   exit(1); 
}

static Color traceray(Vector from, Vector dir,int depth) {
   Color r(0,0,0);
   if(depth++ > MAX_DEPTH) return r;
   intersect in,in2;
   head->findintersect(from,dir,.001,in);
   if(!in.p) return r;
   Vector ip = (dir*in.dist)+from;
   if(!in.gotNormal)
      in.normal  = in.p->findnormal(ip);
   if(in.p->isLum)
      r += in.p->lum;
   if(in.p->isDif) {
    r+=in.p->dif*g.amb;  
   }
   if(in.p->isDif || in.p->isSpec) 
      for(list<lamp*>::iterator li=lamps.begin();li!=lamps.end();li++){
	 lamp *l = *li;
	 Vector dl = l->l-ip;
	 float ld = dl.len();
	 dl = ~dl;
	 in2.p = NULL;
	 head->findintersect(ip,dl,.001,in2);
	 while (in2.p && in2.p->noShadow) 
	    head->findintersect(ip,dl,in2.dist+.001,in2);
	 if((!in2.p) ||(in2.p && in2.dist > ld-.001)) {
	    float cosine = fabs(Dot(in.normal,dl));
	    if(in.p->isDif)
	       r += (l->col*cosine)*in.p->dif/*/((ld*ld))*/;
	    if(in.p->isSpec)
	       r += (l->col*pow(cosine,26))*in.p->spec/*/((ld*ld))*/;
	 }
      }
   if(in.p->isRef) {
      Vector ndir = ~(-(proj(dir,in.normal)*2.0)+dir);
      r += (in.p->ref*traceray(ip,ndir,depth));
   }
   if(in.p->isTrans) 
      r+= in.p->trans*traceray(ip,dir,depth);
   return r;
}


typedef struct tga_header tga_header;
typedef unsigned char uchar;

struct tga_header {
   uchar id;
   uchar comap;
   uchar imgtype;
   uchar indxl,indxhi;
   uchar lonlo,lenhi;
   uchar csize;
   uchar xol,xoh;
   uchar yol,yoh;
   uchar wl,wh;
   uchar hl,hh;
   uchar psize;
   uchar attbits;
};


tga_header default_tga_header = {
   0,0,2,0,0,0,0,
   0,0,0,0,0,0,0,0,0,24,0 
};

int yyparse();
#ifdef YYDEBUG
extern int yydebug;
#endif
int main(int argc,char *argv[]) {
   int i,j;
   Vector xeye,yeye;
#ifdef YYDEBUG
   yydebug = 1;
#endif
   yyparse();
   
   for(list<lamp *>::iterator li=lamps.begin();li!=lamps.end();li++) {
      cerr << "lamp " << (*li)->l << (*li)->col << endl;
   }
   
   if(head)
   cerr << *head;
   else 
      panic("no head!");

   default_tga_header.wl=g.width&0xff;
   default_tga_header.wh=g.width>>8;
   default_tga_header.hl=g.height&0xff;
   default_tga_header.hh=g.height>>8;
   write(STDOUT_FILENO,&default_tga_header,sizeof(default_tga_header));
   xeye = ((~(Cross((Origin-g.camera),Zunit)))/g.zoom);
   yeye = ((~(Cross((Origin-g.camera),xeye)))/g.zoom);
   //xeye = (~((Cross((Origin-g.camera),Zunit))));
   //yeye = (~((Cross((Origin-g.camera),xeye))));
   //xeye *= g.camera.len();
   //yeye *= g.camera.len();
   for(j=g.height-1; j>=0; j--) 
      for(i=0; i<g.width; i++) {
	 Vector dir;
	 float x,y;
	 int  R,G,B;
	 Color c;
	 x = ((float)(i-(g.width>>1)))/(float)(g.width>>1);
	 y = ((float)(j-(g.height>>1)))/(float)(g.height>>1);
	 dir = (((xeye*x)+(yeye*y))-g.camera);
	 c = traceray(g.camera,~dir,0);
	 R =(int)(c.X*255.0);
	 G =(int)(c.Y*255.0);
	 B =(int)(c.Z*255.0);
	 if(R < 0) R = 0;
	 if(R > 255) R = 255;
	 if(G < 0) G = 0;
	 if(G > 255) G = 255;
	 if(B < 0) B = 0;
	 if(B > 255) B = 255;
	 fputc(B, stdout);
	 fputc(G, stdout);
	 fputc(R, stdout);
      }
   return 0;
}
