
%{
#include<stdio.h>
#include<math.h>
#include"engine.h"
#include"multisection.h"
#include"primitive.h"
void yyerror (char *s);
int yylex ();

#define YYDEBUG 1

%}

%union {
float flt;
struct { 
float x,y,z; 
}  vec;
multisection *ms;
sphere *sp;
plane *pl;
object *o;
lamp *l;
}

%token <flt> NUM
%type <flt> exp
%type <vec> vector
%type <o>  object
%type <sp>  sphere_args
%type <sp>  sphere_object
%type <l>   lamp_args
%type <pl>  plane_args
%type <pl>  plane_object
%type <ms>  multisection_object
%type <ms>  ms_list

%left '-' '+'
%left '*' '/'
%left NEG     
%right '^'

%token_table

%%

scene: globals object lamps {head = $2;}

globals: globals global_arg 
	| 

global_arg:	"ambient" vector {g.amb = Vector($2.x,$2.y,$2.z);}
	|	"camera" vector	 {g.camera = Vector($2.x,$2.y,$2.z);}
	|	"dim" exp exp    {g.width = (int)$2; g.height = (int)$3;}
	|	"zoom" exp	 {g.zoom = $2;}


object:   sphere_object  	 {$$ = $1;}
	| plane_object           {$$ = $1;}
	| multisection_object    {$$ = $1;}


sphere_object: "sphere" sphere_args {$$ = $2;$$->post_process();}

sphere_args:  sphere_args "radius" exp {$$ = $1; $$->radius = $3;}
	| sphere_args "location"   vector  {$$=$1; $$->l = Vector($3.x,$3.y,$3.z);}
	| sphere_args "luminance" vector {$$ = $1; $1->lum = Vector($3.x,$3.y,$3.z);$$->isLum=true;}
	| sphere_args "diffuse" vector {$$ = $1; $1->dif = Vector($3.x,$3.y,$3.z);$$->isDif=true;}
	| sphere_args "specular" vector {$$ = $1; $1->spec = Vector($3.x,$3.y,$3.z);$$->isSpec=true;}
	| sphere_args "reflective" vector {$$ = $1; $1->ref = Vector($3.x,$3.y,$3.z);$$->isRef=true;}
	| sphere_args "noshadow" {$$=$1; $$->noShadow = true;}
	|    /* empty */                {$$ = new sphere();}

plane_object: "plane" plane_args  {$$ = $2;$$->post_process();}

plane_args:  plane_args "normal" vector {$$ = $1; $$->normal = Vector($3.x,$3.y,$3.z);}
	| plane_args "location"   vector  {$$=$1; $$->l = Vector($3.x,$3.y,$3.z);}
	| plane_args "luminance" vector {$$ = $1; $1->lum = Vector($3.x,$3.y,$3.z);$$->isLum=true;}
	| plane_args "diffuse" vector {$$ = $1; $1->dif = Vector($3.x,$3.y,$3.z);$$->isDif=true;}
	| plane_args "specular" vector {$$ = $1; $1->spec = Vector($3.x,$3.y,$3.z);$$->isSpec=true;}
	| plane_args "reflective" vector {$$ = $1; $1->ref = Vector($3.x,$3.y,$3.z);$$->isRef=true;}
	| plane_args "noshadow" vector {$$=$1; $$->noShadow = true;}
	|    /* empty */                {$$ = new plane();}

multisection_object: "multisection" exp ms_list "endmultisection" {$$=$3;$$->thresh=(int)$2;}

ms_list: ms_list object {$$ = $1; $$->objs.push_back($2);}
	| /*empty*/ {$$ = new multisection();}


lamps: lamps "lamp" lamp_args  {lamps.push_back($3);}
	| 

lamp_args: lamp_args "location" vector  {$$ = $1; $1->l = Vector($3.x,$3.y,$3.z);} 
	| lamp_args "luminance" vector {$$ = $1; $1->col = Vector($3.x,$3.y,$3.z);}
	| /*empty*/  { $$ = new lamp();}


//vector  { Vector v = Vector($1.x,$1.y,$1.z); cout <<v<< endl;}

vector: exp exp exp      { $$.x = $1; $$.y = $2; $$.z = $3;}
	| exp ','  exp ',' exp   { $$.x = $1; $$.y = $3; $$.z = $5;}
	| '(' vector ')' { $$ = $2;}

exp:      NUM                { $$ = $1;         }
        | exp '+' exp        { $$ = $1 + $3;    }
        | exp '-' exp        { $$ = $1 - $3;    }
        | exp '*' exp        { $$ = $1 * $3;    }
        | exp '/' exp        { $$ = $1 / $3;    }
        | "sin" '(' exp ')'        { $$ = sin($3);}
        | "cos" '(' exp ')'       { $$ = cos($3);}
        | '-' exp  %prec NEG { $$ = -$2;        }
        | exp '^' exp        { $$ = pow ($1, $3); }
        | '(' exp ')'        { $$ = $2;         }
	| "PI"		{ $$ = M_PI;}

%%


int find_tokenstring(char *s) {
   for (int i = 0; i < YYNTOKENS; i++)     {
      if (
		      yytname[i] != 0 
		      && yytname[i][0] == '"'
		      && !strncmp (yytname[i] + 1, s, strlen (s))
		      && yytname[i][strlen (s) + 1] == '"'
		      && yytname[i][strlen (s) + 2] == 0) 
	 return yytoknum[i];
   }
   return 0;
}


void yyerror (char *s) {
  cerr << s << endl;
}
